import torchbearer
from torchbearer import cite
from torchbearer.callbacks import Callback

import torch.nn.init as init

__kaiming__ = """
@inproceedings{he2015delving,
  title={Delving deep into rectifiers: Surpassing human-level performance on imagenet classification},
  author={He, Kaiming and Zhang, Xiangyu and Ren, Shaoqing and Sun, Jian},
  booktitle={Proceedings of the IEEE international conference on computer vision},
  pages={1026--1034},
  year={2015}
}"""

__xavier__ = """
@inproceedings{glorot2010understanding,
  title={Understanding the difficulty of training deep feedforward neural networks},
  author={Glorot, Xavier and Bengio, Yoshua},
  booktitle={Proceedings of the thirteenth international conference on artificial intelligence and statistics},
  pages={249--256},
  year={2010}
}
"""

__lsuv__ = """
@article{mishkin2015all,
  title={All you need is a good init},
  author={Mishkin, Dmytro and Matas, Jiri},
  journal={arXiv preprint arXiv:1511.06422},
  year={2015}
}
"""


class WeightInit(Callback):
    """Base class for weight initialisations. Performs the provided function for each module when on_init is
    called.

    Args:
        initialiser (lambda): a function which initialises an nn.Module **inplace**
        modules (Iterable[nn.Module] or nn.Module, optional): an iterable of nn.Modules or a
            single nn.Module that will have weights initialised, otherwise this is retrieved from the model
        targets (list[String]): A list of lookup strings to match which modules will be initialised

    State Requirements:
        - :attr:`torchbearer.state.MODEL`: Model should have the `modules` method if modules is None
    """
    def __init__(self, initialiser=lambda module: module, modules=None, targets=['Conv', 'Linear', 'Bilinear']):
        self.initialiser = initialiser
        self.modules = modules
        self.targets = targets

    def on_init(self, state):
        if self.modules is None:
            self.modules = state[torchbearer.MODEL].modules()

        for m in self.modules:
            if len(list(filter(lambda target: target in m.__class__.__name__, self.targets))) > 0:
                self.initialiser(m)


@cite(__lsuv__)
class LsuvInit(Callback):
    """Layer-sequential unit-variance (LSUV) initialization as described in
    `All you need is a good init <https://arxiv.org/abs/1511.06422>`_ and
    modified from the code by  `ducha-aiki <https://github.com/ducha-aiki/LSUV-pytorch>`__.
    To be consistent with the paper, LsuvInit should be preceeded by a ZeroBias init on the Linear and Conv layers.


    Example: ::

        >>> import torch
        >>> import torch.nn as nn
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks.init import LsuvInit

        # 100 random data points
        >>> data = torch.rand(100, 3, 5, 5)
        >>> example_batch = data[:3]
        >>> lsuv = LsuvInit(example_batch)

        # Model and trail using lsuv init for some random data
        >>> model = nn.Sequential(nn.Conv2d(3, 1, 3), nn.ReLU())
        >>> trial = Trial(model, callbacks=[lsuv]).with_train_data(data, data+5)

    Args:
        data_item (torch.Tensor): A representative data item to put through the model
        weight_lambda (lambda): A function that takes a module and returns the weight attribute. If none defaults to 
            module.weight.
        needed_std: See `paper <https://arxiv.org/abs/1511.06422>`__, where needed_std is always 1.0
        std_tol: See `paper <https://arxiv.org/abs/1511.06422>`__, Tol_{var}
        max_attempts: See `paper <https://arxiv.org/abs/1511.06422>`__, T_{max}
        do_orthonorm: See `paper <https://arxiv.org/abs/1511.06422>`__, first pre-initialise with orthonormal matricies

    State Requirements:
        - :attr:`torchbearer.state.MODEL`: Model should have the `modules` method if modules is None
    """
    def __init__(self, data_item, weight_lambda=None, needed_std=1.0, std_tol=0.1, max_attempts=10, do_orthonorm=True):
        from torchbearer.callbacks.lsuv import LSUV
        self.lsuv_init = LSUV
        self.data = data_item
        self.needed_std = needed_std
        self.std_tol = std_tol
        self.max_attempts = max_attempts
        self.do_arthonorm = do_orthonorm
        self.weight_lambda = weight_lambda

    def on_init(self, state):
        lsuv = self.lsuv_init()
        state[torchbearer.MODEL] = lsuv.init_model(state[torchbearer.MODEL], self.data, self.weight_lambda, self.needed_std,
                                                  self.std_tol, self.max_attempts, self.do_arthonorm)


@cite(__kaiming__)
class KaimingNormal(WeightInit):
    """Kaiming Normal weight initialisation. Uses `torch.nn.init.kaiming_normal_` on the ``weight`` attribute of the
    filtered modules.

    Example: ::

        >>> import torch
        >>> import torch.nn as nn
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks.init import KaimingNormal

        # 100 random data points
        >>> data = torch.rand(100, 3, 5, 5)
        >>> example_batch = data[:3]
        >>> initialiser = KaimingNormal()

        # Model and trail using kaiming init for some random data
        >>> model = nn.Sequential(nn.Conv2d(3, 1, 3), nn.ReLU())
        >>> trial = Trial(model, callbacks=[initialiser]).with_train_data(data, data+5)

    Args:
        a (int): See `PyTorch kaiming_normal_ <https://pytorch.org/docs/stable/nn.html#torch.nn.init.kaiming_normal_>`_
        mode (str): See `PyTorch kaiming_normal_`_
        nonlinearity (str): See `PyTorch kaiming_normal_`_
        modules (Iterable[nn.Module] or nn.Module, optional): an iterable of nn.Modules or a
            single nn.Module that will have weights initialised, otherwise this is retrieved from the model
        targets (list[String]): A list of lookup strings to match which modules will be initialised

    See:
        `PyTorch kaiming_normal_`_
    """
    def __init__(self, a=0, mode='fan_in', nonlinearity='leaky_relu', modules=None,
                 targets=['Conv', 'Linear', 'Bilinear']):
        def initialiser(module):
            init.kaiming_normal_(module.weight.data, a=a, mode=mode, nonlinearity=nonlinearity)

        super(KaimingNormal, self).__init__(initialiser, modules=modules, targets=targets)


@cite(__kaiming__)
class KaimingUniform(WeightInit):
    """Kaiming Uniform weight initialisation. Uses `torch.nn.init.kaiming_uniform_` on the ``weight`` attribute of the
    filtered modules.

    Example: ::

        >>> import torch
        >>> import torch.nn as nn
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks.init import KaimingUniform

        # 100 random data points
        >>> data = torch.rand(100, 3, 5, 5)
        >>> example_batch = data[:3]
        >>> initialiser = KaimingUniform()

        # Model and trail using kaiming init for some random data
        >>> model = nn.Sequential(nn.Conv2d(3, 1, 3), nn.ReLU())
        >>> trial = Trial(model, callbacks=[initialiser]).with_train_data(data, data+5)

    Args:
        a (int): See `PyTorch kaiming_uniform_ <https://pytorch.org/docs/stable/nn.html#torch.nn.init.kaiming_uniform_>`_
        mode (str): See `PyTorch kaiming_uniform_`_
        nonlinearity (str): See `PyTorch kaiming_uniform_`_
        modules (Iterable[nn.Module] or nn.Module, optional): an iterable of nn.Modules or a
            single nn.Module that will have weights initialised, otherwise this is retrieved from the model
        targets (list[String]): A list of lookup strings to match which modules will be initialised

    See:
        `PyTorch kaiming_uniform_`_
    """
    def __init__(self, a=0, mode='fan_in', nonlinearity='leaky_relu', modules=None,
                 targets=['Conv', 'Linear', 'Bilinear']):
        def initialiser(module):
            init.kaiming_uniform_(module.weight.data, a=a, mode=mode, nonlinearity=nonlinearity)

        super(KaimingUniform, self).__init__(initialiser, modules=modules, targets=targets)


@cite(__xavier__)
class XavierNormal(WeightInit):
    """Xavier Normal weight initialisation. Uses ``torch.nn.init.xavier_normal_`` on the ``weight`` attribute of the
    filtered modules.

    Example: ::

        >>> import torch
        >>> import torch.nn as nn
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks.init import XavierNormal

        # 100 random data points
        >>> data = torch.rand(100, 3, 5, 5)
        >>> example_batch = data[:3]
        >>> initialiser = XavierNormal()

        # Model and trail using Xavier init for some random data
        >>> model = nn.Sequential(nn.Conv2d(3, 1, 3), nn.ReLU())
        >>> trial = Trial(model, callbacks=[initialiser]).with_train_data(data, data+5)

    Args:
        gain (int): See `PyTorch xavier_normal_ <https://pytorch.org/docs/stable/nn.html#torch.nn.init.xavier_normal_>`_
        modules (Iterable[nn.Module] or nn.Module, optional): an iterable of nn.Modules or a
            single nn.Module that will have weights initialised, otherwise this is retrieved from the model
        targets (list[String]): A list of lookup strings to match which modules will be initialised

    See:
        `PyTorch xavier_normal_`_
    """
    def __init__(self, gain=1, modules=None, targets=['Conv', 'Linear', 'Bilinear']):
        def initialiser(module):
            init.xavier_normal_(module.weight.data, gain=gain)

        super(XavierNormal, self).__init__(initialiser, modules=modules, targets=targets)


@cite(__xavier__)
class XavierUniform(WeightInit):
    """Xavier Uniform weight initialisation. Uses ``torch.nn.init.xavier_uniform_`` on the ``weight`` attribute of the
    filtered modules.

    Example: ::

        >>> import torch
        >>> import torch.nn as nn
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks.init import XavierUniform

        # 100 random data points
        >>> data = torch.rand(100, 3, 5, 5)
        >>> example_batch = data[:3]
        >>> initialiser = XavierUniform()

        # Model and trail using Xavier init for some random data
        >>> model = nn.Sequential(nn.Conv2d(3, 1, 3), nn.ReLU())
        >>> trial = Trial(model, callbacks=[initialiser]).with_train_data(data, data+5)

    Args:
        gain (int): See `PyTorch xavier_uniform_ <https://pytorch.org/docs/stable/nn.html#torch.nn.init.xavier_uniform_>`_
        modules (Iterable[nn.Module] or nn.Module, optional): an iterable of nn.Modules or a
            single nn.Module that will have weights initialised, otherwise this is retrieved from the model
        targets (list[String]): A list of lookup strings to match which modules will be initialised

    See:
        `PyTorch xavier_uniform_`_
    """
    def __init__(self, gain=1, modules=None, targets=['Conv', 'Linear', 'Bilinear']):
        def initialiser(module):
            init.xavier_uniform_(module.weight.data, gain=gain)

        super(XavierUniform, self).__init__(initialiser, modules=modules, targets=targets)


class ZeroBias(WeightInit):
    """Zero initialisation for the ``bias`` attributes of filtered modules. This is recommended for use in conjunction
    with weight initialisation schemes.

    Example: ::

        >>> import torch
        >>> import torch.nn as nn
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks.init import ZeroBias

        # 100 random data points
        >>> data = torch.rand(100, 3, 5, 5)
        >>> example_batch = data[:3]
        >>> initialiser = ZeroBias()

        # Model and trail using zero bias init for some random data
        >>> model = nn.Sequential(nn.Conv2d(3, 1, 3), nn.ReLU())
        >>> trial = Trial(model, callbacks=[initialiser]).with_train_data(data, data+5)

    Args:
        modules (Iterable[nn.Module] or nn.Module, optional): an iterable of nn.Modules or a
            single nn.Module that will have weights initialised, otherwise this is retrieved from the model
        targets (list[String]): A list of lookup strings to match which modules will be initialised
    """
    def __init__(self, modules=None, targets=['Conv', 'Linear', 'Bilinear']):
        def initialiser(module):
            module.bias.data.zero_()

        super(ZeroBias, self).__init__(initialiser, modules=modules, targets=targets)



from .version import __version__
from . import magics
from .bases import no_grad, enable_grad, cite, base_closure, Callback, Metric, standard_closure
from .state import *

from . import metrics
from . import callbacks
from .trial import *
from . import cv_utils



from unittest import TestCase

import torchbearer
from torchbearer.callbacks import TerminateOnNaN


class TestTerminateOnNaN(TestCase):
    def test_should_terminate(self):
        state = {
            torchbearer.EPOCH: 0,
            torchbearer.BATCH: 0,
            torchbearer.STOP_TRAINING: False,
            torchbearer.METRICS: {'running_loss': float('nan')}
        }

        terminator = TerminateOnNaN()

        terminator.on_step_training(state)
        self.assertTrue(state[torchbearer.STOP_TRAINING])
        terminator.on_step_validation(state)
        self.assertTrue(state[torchbearer.STOP_TRAINING])
        terminator.on_end_epoch(state)
        self.assertTrue(state[torchbearer.STOP_TRAINING])

    def test_should_not_terminate(self):
        state = {
            torchbearer.EPOCH: 0,
            torchbearer.BATCH: 0,
            torchbearer.STOP_TRAINING: False,
            torchbearer.METRICS: {'running_loss': 0.01}
        }

        terminator = TerminateOnNaN()

        terminator.on_step_training(state)
        self.assertFalse(state[torchbearer.STOP_TRAINING])
        terminator.on_step_validation(state)
        self.assertFalse(state[torchbearer.STOP_TRAINING])
        terminator.on_end_epoch(state)
        self.assertFalse(state[torchbearer.STOP_TRAINING])

    def test_monitor_should_terminate(self):
        state = {
            torchbearer.EPOCH: 0,
            torchbearer.BATCH: 0,
            torchbearer.STOP_TRAINING: False,
            torchbearer.METRICS: {'running_loss': 0.001, 'test_metric': float('nan')}
        }

        terminator = TerminateOnNaN(monitor='test_metric')

        terminator.on_step_training(state)
        self.assertTrue(state[torchbearer.STOP_TRAINING])
        terminator.on_step_validation(state)
        self.assertTrue(state[torchbearer.STOP_TRAINING])
        terminator.on_end_epoch(state)
        self.assertTrue(state[torchbearer.STOP_TRAINING])

    def test_monitor_should_not_terminate(self):
        state = {
            torchbearer.EPOCH: 0,
            torchbearer.BATCH: 0,
            torchbearer.STOP_TRAINING: False,
            torchbearer.METRICS: {'running_loss': 0.01, 'test_metric': 0.001}
        }

        terminator = TerminateOnNaN(monitor='test_metric')

        terminator.on_step_training(state)
        self.assertFalse(state[torchbearer.STOP_TRAINING])
        terminator.on_step_validation(state)
        self.assertFalse(state[torchbearer.STOP_TRAINING])
        terminator.on_end_epoch(state)
        self.assertFalse(state[torchbearer.STOP_TRAINING])

    def test_not_found_metric(self):
        state = {
            torchbearer.EPOCH: 0,
            torchbearer.BATCH: 0,
            torchbearer.STOP_TRAINING: False,
            torchbearer.METRICS: {'running_loss': 0.001, 'test_metric': float('nan')}
        }

        terminator = TerminateOnNaN(monitor='test')

        terminator.on_step_training(state)
        self.assertFalse(state[torchbearer.STOP_TRAINING])
        terminator.on_step_validation(state)
        self.assertFalse(state[torchbearer.STOP_TRAINING])
        terminator.on_end_epoch(state)
        self.assertFalse(state[torchbearer.STOP_TRAINING])



from unittest import TestCase

import torch

import torchbearer
from torchbearer.callbacks import LabelSmoothingRegularisation


class TestLabelSmoothingRegularisation(TestCase):
    def test_with_long_target(self):
        callback = LabelSmoothingRegularisation(epsilon=0.4, classes=4)
        state = {torchbearer.TARGET: torch.Tensor([1, 3]).long()}

        target = torch.Tensor([
            [0.1, 0.7, 0.1, 0.1],
            [0.1, 0.1, 0.1, 0.7]
        ])

        callback.on_sample(state)
        self.assertTrue(((state[torchbearer.TARGET] - target).abs() < 0.00001).all())

    def test_with_hot_target(self):
        callback = LabelSmoothingRegularisation(epsilon=0.4)
        state = {torchbearer.TARGET: torch.Tensor([
            [0, 1, 0, 0],
            [0, 0, 0, 1]
        ]).long()}

        target = torch.Tensor([
            [0.1, 0.7, 0.1, 0.1],
            [0.1, 0.1, 0.1, 0.7]
        ])

        callback.on_sample(state)
        self.assertTrue(((state[torchbearer.TARGET] - target).abs() < 0.00001).all())



import sys

import torchbearer

from torchbearer.callbacks import Callback
import csv


class CSVLogger(Callback):
    """Callback to log metrics to a given csv file.

    Example: ::

        >>> from torchbearer.callbacks import CSVLogger
        >>> from torchbearer import Trial
        >>> import torch

        # Example Trial (without optimiser or loss criterion) which writes metrics to a csv file appending to previous content
        >>> logger = CSVLogger('my_path.pt', separator=',', append=True)
        >>> trial = Trial(None, callbacks=[logger], metrics=['acc'])

    Args:
        filename (str): The name of the file to output to
        separator (str): The delimiter to use (e.g. comma, tab etc.)
        batch_granularity (bool): If True, write on each batch, else on each epoch
        write_header (bool): If True, write the CSV header at the beginning of training
        append (bool): If True, append to the file instead of replacing it

    State Requirements:
        - :attr:`torchbearer.state.EPOCH`: State should have the current epoch stored
        - :attr:`torchbearer.state.METRICS`: Metrics dictionary should exist
        - :attr:`torchbearer.state.BATCH`: State should have the current batch stored if using `batch_granularity`
    """

    def __init__(self, filename, separator=',', batch_granularity=False, write_header=True, append=False):

        super(CSVLogger, self).__init__()
        self.batch_granularity = batch_granularity
        self.filename = filename
        self.separator = separator
        if append:
            filemode = 'a'
        else:
            filemode = 'w'

        if sys.version_info[0] < 3:
            filemode += 'b'
            self.csvfile = open(self.filename, filemode)
        else:
            self.csvfile = open(self.filename, filemode, newline='')
        self.write_header = write_header

    def on_step_training(self, state):
        super(CSVLogger, self).on_step_training(state)
        if self.batch_granularity:
            self._write_to_dict(state)

    def on_end_epoch(self, state):
        super(CSVLogger, self).on_end_training(state)
        self._write_to_dict(state)

    def on_end(self, state):
        super(CSVLogger, self).on_end(state)
        self.csvfile.close()

    def _write_to_dict(self, state):
        fields = self._get_field_dict(state)
        self.writer = csv.DictWriter(self.csvfile, fieldnames=fields.keys(), delimiter=self.separator)

        if self.write_header:
            self.writer.writeheader()
            self.write_header = False

        self.writer.writerow(fields)
        self.csvfile.flush()

    def _get_field_dict(self, state):
        fields = {'epoch': state[torchbearer.EPOCH]}

        if self.batch_granularity:
            fields.update({'batch': state[torchbearer.BATCH]})

        fields.update(state[torchbearer.METRICS])

        return fields



import torchbearer

import torch

from torchbearer.callbacks.callbacks import Callback
import os
import warnings


class _Checkpointer(Callback):
    def __init__(self, fileformat, save_model_params_only=False, pickle_module=torch.serialization.pickle, pickle_protocol=torch.serialization.DEFAULT_PROTOCOL):
        super(_Checkpointer, self).__init__()
        self.fileformat = fileformat

        self.pickle_module = pickle_module
        self.pickle_protocol = pickle_protocol

        self.save_model_params_only = save_model_params_only

        self.most_recent = None

        if fileformat.__contains__(os.sep) and not os.path.exists(os.path.dirname(fileformat)):
            os.makedirs(os.path.dirname(fileformat))

    def save_checkpoint(self, model_state, overwrite_most_recent=False):
        state = {}
        state.update(model_state)
        state.update(model_state[torchbearer.METRICS])

        string_state = {str(key): state[key] for key in state.keys()}
        filepath = self.fileformat.format(**string_state)

        if self.most_recent is not None and overwrite_most_recent:
            try:
                os.remove(self.most_recent)
            except OSError:
                warnings.warn('Failed to delete old file. Are you running two checkpointers with the same filename?')

        if self.save_model_params_only:
            torch.save(model_state[torchbearer.MODEL].state_dict(), filepath, pickle_module=self.pickle_module,
                       pickle_protocol=self.pickle_protocol)
        else:
            torch.save(model_state[torchbearer.SELF].state_dict(), filepath, pickle_module=self.pickle_module,
                       pickle_protocol=self.pickle_protocol)

        self.most_recent = filepath


def ModelCheckpoint(filepath='model.{epoch:02d}-{val_loss:.2f}.pt', save_model_params_only=False,
        monitor='val_loss', save_best_only=False, mode='auto', period=1, min_delta=0):
    """Save the model after every epoch. `filepath` can contain named formatting options, which will be filled any
    values from state. For example: if `filepath` is `weights.{epoch:02d}-{val_loss:.2f}`, then the model checkpoints
    will be saved with the epoch number and the validation loss in the filename. The torch :class:`.Trial` will be
    saved to filename.


    Example: ::

        >>> from torchbearer.callbacks import ModelCheckpoint
        >>> from torchbearer import Trial
        >>> import torch

        # Example Trial (without optimiser or loss criterion) which uses this checkpointer
        >>> model = torch.nn.Linear(1,1)
        >>> checkpoint = ModelCheckpoint('my_path.pt', monitor='val_acc', mode='max')
        >>> trial = Trial(model, callbacks=[checkpoint], metrics=['acc'])

    Args:
        filepath (str): Path to save the model file
        save_model_params_only (bool): If `save_model_params_only=True`, only model parameters will be saved so that
            the results can be loaded into a PyTorch nn.Module. The other option, `save_model_params_only=False`,
            should be used only if the results will be loaded into a Torchbearer Trial object later.
        monitor (str): Quantity to monitor
        save_best_only (bool): If `save_best_only=True`, the latest best model according to the quantity
            monitored will not be overwritten
        mode (str): One of {auto, min, max}. If `save_best_only=True`, the decision to overwrite the current
            save file is made based on either the maximization or the minimization of the monitored quantity. For
            `val_acc`, this should be `max`, for `val_loss` this should be `min`, etc. In `auto` mode, the direction is
            automatically inferred from the name of the monitored quantity.
        period (int): Interval (number of epochs) between checkpoints
        min_delta (float): If `save_best_only=True`, this is the minimum improvement required to trigger a save

    State Requirements:
        - :attr:`torchbearer.state.MODEL`: Model should have the `state_dict` method
        - :attr:`torchbearer.state.METRICS`: Metrics dictionary should exist
        - :attr:`torchbearer.state.SELF`: Self should be the :attr:`torchbearer.Trial` which is running this callback
    """
    if save_best_only:
        check = Best(filepath, save_model_params_only, monitor, mode, period, min_delta)
    else:
        check = Interval(filepath, save_model_params_only, period)

    return check


class MostRecent(_Checkpointer):
    """Model checkpointer which saves the most recent model to a given filepath. `filepath` can contain named
    formatting options, which will be filled any values from state. For example: if `filepath` is
    `weights.{epoch:02d}-{val_loss:.2f}`, then the model checkpoints will be saved with the epoch number and the
    validation loss in the filename.

    Example: ::

        >>> from torchbearer.callbacks import MostRecent
        >>> from torchbearer import Trial
        >>> import torch

        # Example Trial (without optimiser or loss criterion) which uses this checkpointer
        >>> model = torch.nn.Linear(1,1)
        >>> checkpoint = MostRecent('my_path.pt')
        >>> trial = Trial(model, callbacks=[checkpoint], metrics=['acc'])

    Args:
        filepath (str): Path to save the model file
        save_model_params_only (bool): If `save_model_params_only=True`, only model parameters will be saved so that
            the results can be loaded into a PyTorch nn.Module. The other option, `save_model_params_only=False`,
            should be used only if the results will be loaded into a Torchbearer Trial object later.
        pickle_module (module): The pickle module to use, default is 'torch.serialization.pickle'
        pickle_protocol (int): The pickle protocol to use, default is 'torch.serialization.DEFAULT_PROTOCOL'

    State Requirements:
        - :attr:`torchbearer.state.MODEL`: Model should have the `state_dict` method
        - :attr:`torchbearer.state.METRICS`: Metrics dictionary should exist
        - :attr:`torchbearer.state.SELF`: Self should be the :attr:`torchbearer.Trial` which is running this callback
    """

    def __init__(self, filepath='model.{epoch:02d}-{val_loss:.2f}.pt', save_model_params_only=False,
                 pickle_module=torch.serialization.pickle, pickle_protocol=torch.serialization.DEFAULT_PROTOCOL):

        super(MostRecent, self).__init__(filepath, save_model_params_only=save_model_params_only,
                                         pickle_module=pickle_module, pickle_protocol=pickle_protocol)
        self.filepath = filepath

    def on_checkpoint(self, state):
        super(MostRecent, self).on_end_epoch(state)
        self.save_checkpoint(state, overwrite_most_recent=True)


class Best(_Checkpointer):
    """Model checkpointer which saves the best model according to the given configurations. `filepath` can contain
    named formatting options, which will be filled any values from state. For example: if `filepath` is
    `weights.{epoch:02d}-{val_loss:.2f}`, then the model checkpoints will be saved with the epoch number and the
    validation loss in the filename.

    Example: ::

        >>> from torchbearer.callbacks import Best
        >>> from torchbearer import Trial
        >>> import torch

        # Example Trial (without optimiser or loss criterion) which uses this checkpointer
        >>> model = torch.nn.Linear(1,1)
        >>> checkpoint = Best('my_path.pt', monitor='val_acc', mode='max')
        >>> trial = Trial(model, callbacks=[checkpoint], metrics=['acc'])

    Args:
        filepath (str): Path to save the model file
        save_model_params_only (bool): If `save_model_params_only=True`, only model parameters will be saved so that
            the results can be loaded into a PyTorch nn.Module. The other option, `save_model_params_only=False`,
            should be used only if the results will be loaded into a Torchbearer Trial object later.
        monitor (str): Quantity to monitor
        mode (str): One of {auto, min, max}. If `save_best_only=True`, the decision to overwrite the current save file
            is made based on either the maximization or the minimization of the monitored quantity. For `val_acc`, this
            should be `max`, for `val_loss` this should be `min`, etc. In `auto` mode, the direction is automatically
            inferred from the name of the monitored quantity.
        period (int): Interval (number of epochs) between checkpoints
        min_delta (float): If `save_best_only=True`, this is the minimum improvement required to trigger a save
        pickle_module (module): The pickle module to use, default is 'torch.serialization.pickle'
        pickle_protocol (int): The pickle protocol to use, default is 'torch.serialization.DEFAULT_PROTOCOL'

    State Requirements:
        - :attr:`torchbearer.state.MODEL`: Model should have the `state_dict` method
        - :attr:`torchbearer.state.METRICS`: Metrics dictionary should exist, with the `monitor` key populated
        - :attr:`torchbearer.state.SELF`: Self should be the :attr:`torchbearer.Trial` which is running this callback
    """

    def __init__(self, filepath='model.{epoch:02d}-{val_loss:.2f}.pt', save_model_params_only=False, monitor='val_loss',
                 mode='auto', period=1, min_delta=0, pickle_module=torch.serialization.pickle,
                 pickle_protocol=torch.serialization.DEFAULT_PROTOCOL):

        super(Best, self).__init__(filepath, save_model_params_only=save_model_params_only,
                                   pickle_module=pickle_module, pickle_protocol=pickle_protocol)
        self.min_delta = min_delta
        self.mode = mode
        self.monitor = monitor
        self.period = period
        self.epochs_since_last_save = 0

        if self.mode not in ['min', 'max']:
            if 'acc' in self.monitor:
                self.mode = 'max'
            else:
                self.mode = 'min'

        if self.mode == 'min':
            self.min_delta *= -1
            self.monitor_op = lambda x1, x2: (x1-self.min_delta) < x2
        elif self.mode == 'max':
            self.min_delta *= 1
            self.monitor_op = lambda x1, x2: (x1-self.min_delta) > x2

        self.best = None

    def state_dict(self):
        state_dict = super(Best, self).state_dict()
        state_dict['epochs'] = self.epochs_since_last_save
        state_dict['best'] = self.best

        return state_dict

    def load_state_dict(self, state_dict):
        super(Best, self).load_state_dict(state_dict)
        self.epochs_since_last_save = state_dict['epochs']
        self.best = state_dict['best']

        return self

    def on_start(self, state):
        if self.best is None:
            self.best = float('inf') if self.mode == 'min' else -float('inf')

    def on_checkpoint(self, state):
        super(Best, self).on_end_epoch(state)
        self.epochs_since_last_save += 1
        if self.epochs_since_last_save >= self.period:
            self.epochs_since_last_save = 0

            current = state[torchbearer.METRICS][self.monitor]

            if self.monitor_op(current, self.best):
                self.best = current
                self.save_checkpoint(state, overwrite_most_recent=True)


class Interval(_Checkpointer):
    """Model checkpointer which which saves the model every 'period' epochs to the given filepath. `filepath` can
    contain named formatting options, which will be filled any values from state. For example: if `filepath` is
    `weights.{epoch:02d}-{val_loss:.2f}`, then the model checkpoints will be saved with the epoch number and the
    validation loss in the filename.

    Example: ::

        >>> from torchbearer.callbacks import Interval
        >>> from torchbearer import Trial
        >>> import torch

        # Example Trial (without optimiser or loss criterion) which uses this checkpointer
        >>> model = torch.nn.Linear(1,1)
        >>> checkpoint = Interval('my_path.pt', period=100, on_batch=True)
        >>> trial = Trial(model, callbacks=[checkpoint], metrics=['acc'])

    Args:
        filepath (str): Path to save the model file
        save_model_params_only (bool): If `save_model_params_only=True`, only model parameters will be saved so that
            the results can be loaded into a PyTorch nn.Module. The other option, `save_model_params_only=False`,
            should be used only if the results will be loaded into a Torchbearer Trial object later.
        period (int): Interval (number of steps) between checkpoints
        on_batch (bool): If true step each batch, if false step each epoch.
        period (int): Interval (number of epochs) between checkpoints
        pickle_module (module): The pickle module to use, default is 'torch.serialization.pickle'
        pickle_protocol (int): The pickle protocol to use, default is 'torch.serialization.DEFAULT_PROTOCOL'

    State Requirements:
        - :attr:`torchbearer.state.MODEL`: Model should have the `state_dict` method
        - :attr:`torchbearer.state.METRICS`: Metrics dictionary should exist
        - :attr:`torchbearer.state.SELF`: Self should be the :attr:`torchbearer.Trial` which is running this callback
    """

    def __init__(self, filepath='model.{epoch:02d}-{val_loss:.2f}.pt', save_model_params_only=False, period=1, on_batch=False, pickle_module=torch.serialization.pickle, pickle_protocol=torch.serialization.DEFAULT_PROTOCOL):

        super(Interval, self).__init__(filepath, save_model_params_only=save_model_params_only,
                                       pickle_module=pickle_module, pickle_protocol=pickle_protocol)
        self.period = period
        self.epochs_since_last_save = 0

        if on_batch:
            self.on_step_training = self.on_checkpoint
            self.on_checkpoint = lambda _: None

    def state_dict(self):
        state_dict = super(Interval, self).state_dict()
        state_dict['epochs'] = self.epochs_since_last_save

        return state_dict

    def load_state_dict(self, state_dict):
        super(Interval, self).load_state_dict(state_dict)
        self.epochs_since_last_save = state_dict['epochs']

        return self

    def on_checkpoint(self, state):
        super(Interval, self).on_end_epoch(state)

        self.epochs_since_last_save += 1
        if self.epochs_since_last_save >= self.period:
            self.epochs_since_last_save = 0
            self.save_checkpoint(state)



from __future__ import print_function

import torch

import torchbearer
from torchbearer import Callback
from torchbearer.bases import cite
from torchbearer.metrics import EpochLambda, MetricList


pycm_bib = """
@article{Haghighi2018,
  doi = {10.21105/joss.00729},
  url = {https://doi.org/10.21105/joss.00729},
  year  = {2018},
  month = {may},
  publisher = {The Open Journal},
  volume = {3},
  number = {25},
  pages = {729},
  author = {Sepand Haghighi and Masoomeh Jasemi and Shaahin Hessabi and Alireza Zolanvari},
  title = {{PyCM}: Multiclass confusion matrix library in Python},
  journal = {Journal of Open Source Software}
}
"""


def _to_pyplot(normalize=False, title='Confusion matrix', cmap=None):
    """
    This function modified to plots the ConfusionMatrix object.
    Normalization can be applied by setting `normalize=True`.

    Code Reference :
    http://scikit-learn.org/stable/auto_examples/model_selection/plot_confusion_matrix.html

    """
    import matplotlib.pyplot as plt
    import itertools
    import numpy as np

    if cmap is None:
        cmap = plt.cm.Blues

    def handler(cm, state):
        string_state = {str(key): state[key] for key in state.keys()}
        plt_cm = []
        for i in cm.classes:
            row = []
            for j in cm.classes:
                row.append(cm.table[i][j])
            plt_cm.append(row)
        plt_cm = np.array(plt_cm)
        if normalize:
            plt_cm = plt_cm.astype('float') / plt_cm.sum(axis=1)[:, np.newaxis]
        plt.imshow(plt_cm, interpolation='nearest', cmap=cmap)
        plt.title(title.format(**string_state))
        plt.colorbar()
        tick_marks = np.arange(len(cm.classes))
        plt.xticks(tick_marks, cm.classes, rotation=45)
        plt.yticks(tick_marks, cm.classes)

        fmt = '.2f' if normalize else 'd'
        thresh = plt_cm.max() / 2.
        for i, j in itertools.product(range(plt_cm.shape[0]), range(plt_cm.shape[1])):
            plt.text(j, i, format(plt_cm[i, j], fmt),
                     horizontalalignment="center",
                     color="white" if plt_cm[i, j] > thresh else "black")

        plt.tight_layout()
        plt.ylabel('Actual')
        plt.xlabel('Predict')
        plt.show()
    return handler


@cite(pycm_bib)
class PyCM(Callback):
    """Create confusion matrices using the `PyCM library <https://github.com/sepandhaghighi/pycm>`_.

    Args:
        kwargs: Additional keyword args to pass to the `ConfusionMatrix` call
    """
    def __init__(self, **kwargs):
        import sys
        if sys.version_info[0] < 3:
            raise Exception('PyCM requires Python>=3!')
        self._handlers = []

        self.kwargs = kwargs

    def _add_metric(self, state):
        from pycm import ConfusionMatrix

        def make_cm(y_pred, y_true):
            _, y_pred = torch.max(y_pred, 1)
            cm = ConfusionMatrix(y_true.cpu().numpy(), y_pred.cpu().numpy(), **self.kwargs)
            for handler in self._handlers:
                handler(cm, state)

        my_metric = EpochLambda('pycm', make_cm, False)
        my_metric.reset(state)
        state[torchbearer.METRIC_LIST] = MetricList([state[torchbearer.METRIC_LIST], my_metric])

    def on_train(self):
        """Process this callback for training batches

        Returns:
            PyCM: self
        """
        _old_start_training = self.on_start_training

        def wrapper(state):
            _old_start_training(state)
            self._add_metric(state)

        self.on_start_training = wrapper
        return self

    def on_val(self):
        """Process this callback for validation batches

        Returns:
            PyCM: self
        """
        _old_start_validation = self.on_start_validation

        def wrapper(state):
            _old_start_validation(state)
            if state[torchbearer.DATA] is torchbearer.VALIDATION_DATA:
                self._add_metric(state)

        self.on_start_validation = wrapper
        return self

    def on_test(self):
        """Process this callback for test batches

        Returns:
            PyCM: self
        """
        _old_start_validation = self.on_start_validation

        def wrapper(state):
            _old_start_validation(state)
            if state[torchbearer.DATA] is torchbearer.TEST_DATA:
                self._add_metric(state)

        self.on_start_validation = wrapper
        return self

    def with_handler(self, handler):
        """Append the given output handler to the list of handlers

        Args:
            handler: A function of confusion and matrix and state

        Returns:
            PyCM: self
        """
        self._handlers.append(handler)
        return self

    def to_state(self, key):
        """Send `ConfusionMatrix` objects from this callback to the given state key

        Args:
            key (StateKey): The key to store the confusion matrix in

        Returns:
            PyCM: self
        """
        def handler(cm, state):
            state[key] = cm
        return self.with_handler(handler)

    def to_console(self):
        """Print `ConfusionMatrix` objects from this callback to the console

        Returns:
            PyCM: self
        """
        return self.with_handler(lambda cm, _: print(cm))

    def to_pycm_file(self, filename, address=True, overall_param=None, class_param=None, class_name=None):
        """Save `ConfusionMatrix` objects from this callback to `.pycm` files

        Args:
            filename (str): The name of the file, will be formatted with state to create unique filenames if desired

        See:
            `PyCM Source <https://github.com/sepandhaghighi/pycm/blob/master/pycm/pycm_obj.py>`_

        Returns:
            PyCM: self
        """
        def handler(cm, state):
            string_state = {str(key): state[key] for key in state.keys()}
            cm.save_stat(filename.format(**string_state), address=address, overall_param=overall_param,
                         class_param=class_param, class_name=class_name)
        return self.with_handler(handler)

    def to_html_file(self, filename, address=True, overall_param=None, class_param=None, class_name=None,
                     color=(0, 0, 0), normalize=False):
        """Save `ConfusionMatrix` objects from this callback to `.html` files

        Args:
            filename (str): The name of the file, will be formatted with state to create unique filenames if desired

        See:
            `PyCM Source <https://github.com/sepandhaghighi/pycm/blob/master/pycm/pycm_obj.py>`_

        Returns:
            PyCM: self
        """
        def handler(cm, state):
            string_state = {str(key): state[key] for key in state.keys()}
            cm.save_html(filename.format(**string_state), address=address, overall_param=overall_param,
                         class_param=class_param, class_name=class_name, color=color, normalize=normalize)
        return self.with_handler(handler)

    def to_csv_file(self, filename, address=True, overall_param=None, class_param=None, class_name=None,
                    matrix_save=True, normalize=False):
        """Save `ConfusionMatrix` objects from this callback to `.csv` files

        Args:
            filename (str): The name of the file, will be formatted with state to create unique filenames if desired

        See:
            `PyCM Source <https://github.com/sepandhaghighi/pycm/blob/master/pycm/pycm_obj.py>`_

        Returns:
            PyCM: self
        """
        def handler(cm, state):
            string_state = {str(key): state[key] for key in state.keys()}
            cm.save_csv(filename.format(**string_state), address=address, overall_param=overall_param,
                        class_param=class_param, class_name=class_name, matrix_save=matrix_save, normalize=normalize)
        return self.with_handler(handler)

    def to_obj_file(self, filename, address=True, save_stat=False, save_vector=True):
        """Save `ConfusionMatrix` objects from this callback to `.obj` files

        Args:
            filename (str): The name of the file, will be formatted with state to create unique filenames if desired

        See:
            `PyCM Source <https://github.com/sepandhaghighi/pycm/blob/master/pycm/pycm_obj.py>`_

        Returns:
            PyCM: self
        """
        def handler(cm, state):
            string_state = {str(key): state[key] for key in state.keys()}
            cm.save_obj(filename.format(**string_state), address=address, save_stat=save_stat, save_vector=save_vector)
        return self.with_handler(handler)

    def to_pyplot(self, normalize=False, title='Confusion matrix', cmap=None):
        """Plot `ConfusionMatrix` objects from this callback with `matplotlib.pyplot`

        Args:
            normalize (bool): If true, normalize the confusion matrix
            title (str): Title to use for the plot, will be formatted with state to create unique titles if desired
            cmap: Colour map object to use for the plot, defaults to `plt.cm.Blues` if None

        See:
            `PyCM Source <https://github.com/sepandhaghighi/pycm/blob/master/pycm/pycm_obj.py>`_

        Returns:
            PyCM: self
        """
        return self.with_handler(_to_pyplot(normalize=normalize, title=title, cmap=cmap))



from unittest import TestCase

import torch
import torch.nn as nn
from mock import Mock, patch

import torchbearer
from torchbearer.callbacks import WeightDecay, L1WeightDecay, L2WeightDecay


class TestWeightDecay(TestCase):

    def test_single_parameter(self):
        model = nn.Sequential(nn.Conv2d(3, 3, 3))
        model.parameters = Mock(return_value=[nn.Parameter(torch.Tensor([2.0]))])
        state = {
            torchbearer.MODEL: model,
            torchbearer.LOSS: 1
        }

        decay = WeightDecay(1, 1.0)
        decay.on_start(state)
        decay.on_criterion(state)

        self.assertTrue(state[torchbearer.LOSS].item() == 3)

    def test_multiple_parameters(self):
        model = nn.Sequential(nn.Conv2d(3, 3, 3))
        model.parameters = Mock(return_value=[nn.Parameter(torch.Tensor([2.0, 1.0]))])
        state = {
            torchbearer.MODEL: model,
            torchbearer.LOSS: 1
        }

        decay = WeightDecay(1, 1.0)
        decay.on_start(state)
        decay.on_criterion(state)

        self.assertTrue(state[torchbearer.LOSS].item() == 4)

    def test_rate(self):
        model = nn.Sequential(nn.Conv2d(3, 3, 3))
        model.parameters = Mock(return_value=[nn.Parameter(torch.Tensor([2.0]))])
        state = {
            torchbearer.MODEL: model,
            torchbearer.LOSS: 1
        }

        decay = WeightDecay(0.5, 1.0)
        decay.on_start(state)
        decay.on_criterion(state)

        self.assertTrue(state[torchbearer.LOSS].item() == 2)

    def test_pow(self):
        model = nn.Sequential(nn.Conv2d(3, 3, 3))
        model.parameters = Mock(return_value=[nn.Parameter(torch.Tensor([2.0]))])
        state = {
            torchbearer.MODEL: model,
            torchbearer.LOSS: 1
        }

        decay = WeightDecay(2, 2)
        decay.on_start(state)
        decay.on_criterion(state)

        self.assertTrue(state[torchbearer.LOSS].item() == 5)

    def test_given_params(self):
        model = nn.Sequential(nn.Conv2d(3, 3, 3))
        model.parameters = Mock(return_value=-1)
        state = {
            torchbearer.MODEL: model,
            torchbearer.LOSS: 1
        }

        decay = WeightDecay(2, 2, params=model.parameters())
        decay.on_start(state)

        self.assertTrue(decay.params == -1)

    @patch('torchbearer.callbacks.weight_decay.WeightDecay.__init__')
    def test_L1(self, mock_wd):
        model = nn.Sequential(nn.Conv2d(3, 3, 3))
        model.parameters = Mock(return_value=-1)
        state = {
            torchbearer.MODEL: model,
            torchbearer.LOSS: 1
        }

        decay = L1WeightDecay()
        self.assertTrue(mock_wd.call_args[1]['p'] == 1)

    @patch('torchbearer.callbacks.weight_decay.WeightDecay.__init__')
    def test_L2(self, mock_wd):
        model = nn.Sequential(nn.Conv2d(3, 3, 3))
        model.parameters = Mock(return_value=-1)
        state = {
            torchbearer.MODEL: model,
            torchbearer.LOSS: 1
        }

        decay = L2WeightDecay()
        self.assertTrue(mock_wd.call_args[1]['p'] == 2)



"""
    .. autoclass:: LR()
"""
import torchbearer
from .metrics import AdvancedMetric
from .decorators import default_for_key, to_dict

old_super = super


def super(_, obj):
    return old_super(obj.__class__, obj)


def _get_lr(optimizer):
    lrs = []
    for group in optimizer.param_groups:
        lrs.append(group['lr'])

    if len(lrs) == 1:
        return lrs[0]
    return lrs


@default_for_key('lr')
@to_dict
class LR(AdvancedMetric):
    """Returns the learning rate/s from the optimizer group/s. Use this to log the current LR when using decay.
    Default for key 'lr'

    State Requirements:
        - :attr:`torchbearer.state.OPTIMIZER`: The optimizer in state will be used to retrieve the learning rate.
    """

    def __init__(self):
        super(LR, self).__init__('lr')

    def process_train(self, *args):
        state = args[0]
        return _get_lr(state[torchbearer.OPTIMIZER])

    def process_final_train(self, *args):
        state = args[0]
        return _get_lr(state[torchbearer.OPTIMIZER])



from unittest import TestCase

from mock import MagicMock, patch, ANY

import torch

import torchbearer
import torchbearer.callbacks.imaging as imaging

import matplotlib.pyplot as plt  # Import so that it can be mocked
plt.ioff()


class TestHandlers(TestCase):
    @patch('PIL.Image')
    def test_to_file(self, pil):
        handler = imaging.imaging._to_file('test')
        mock = MagicMock()
        handler(mock, 0, '')
        mock.mul.assert_called_once_with(255)
        mock.mul().clamp.assert_called_once_with(0, 255)
        self.assertTrue(mock.mul().clamp().byte.call_count == 1)
        mock.mul().clamp().byte().permute.assert_called_once_with(1, 2, 0)
        self.assertTrue(mock.mul().clamp().byte().permute().cpu.call_count == 1)
        self.assertTrue(mock.mul().clamp().byte().permute().cpu().numpy.call_count == 1)

        pil.fromarray.assert_called_once_with(mock.mul().clamp().byte().permute().cpu().numpy())
        pil.fromarray().save.assert_called_once_with('test')

    @patch('matplotlib.pyplot')
    def test_to_pyplot(self, plt):
        handler = imaging.imaging._to_pyplot(title='test')
        mock = MagicMock()
        handler(mock, 0, '')
        mock.mul.assert_called_once_with(255)
        mock.mul().clamp.assert_called_once_with(0, 255)
        self.assertTrue(mock.mul().clamp().byte.call_count == 1)
        mock.mul().clamp().byte().permute.assert_called_once_with(1, 2, 0)
        self.assertTrue(mock.mul().clamp().byte().permute().cpu.call_count == 1)
        self.assertTrue(mock.mul().clamp().byte().permute().cpu().numpy.call_count == 1)

        plt.title.assert_called_once_with('test')

        plt.imshow.assert_called_once_with(mock.mul().clamp().byte().permute().cpu().numpy())
        plt.axis.assert_called_once_with('off')
        self.assertTrue(plt.show.call_count == 1)

    @patch('torchbearer.callbacks.tensor_board')
    def test_to_tensorboard(self, tboard):
        handler = imaging.imaging._to_tensorboard('test', log_dir='./logs', comment='comment')
        image = MagicMock()
        handler(image, 0, {torchbearer.EPOCH: 1})
        image.clamp.assert_called_once_with(0, 1)
        tboard.get_writer.assert_called_once_with('./logs/comment', imaging.imaging._to_tensorboard)
        tboard.get_writer().add_image.assert_called_once_with('test', image.clamp(), 1)
        tboard.close_writer.assert_called_once_with('./logs/comment', imaging.imaging._to_tensorboard)

    @patch('torchbearer.callbacks.tensor_board')
    def test_to_visdom(self, tboard):
        handler = imaging.imaging._to_visdom('test', log_dir='./logs', comment='comment', visdom_params='test_params')
        image = MagicMock()
        handler(image, 0, {torchbearer.EPOCH: 1})
        image.clamp.assert_called_once_with(0, 1)
        tboard.get_writer.assert_called_once_with('./logs/comment', imaging.imaging._to_visdom, visdom=True, visdom_params='test_params')
        tboard.get_writer().add_image.assert_called_once_with('test_1', image.clamp(), 1)
        tboard.close_writer.assert_called_once_with('./logs/comment', imaging.imaging._to_visdom)


class TestImagingCallback(TestCase):
    def test_transform(self):
        callback = imaging.ImagingCallback()
        self.assertTrue(callback.transform('test') is 'test')

        callback = imaging.ImagingCallback(transform=lambda _: 'test')
        self.assertTrue(callback.transform('something else') is 'test')

    @patch('torchbearer.callbacks.imaging.imaging._cache_images')
    def test_cache(self, mock_cache_images):
        callback = imaging.ImagingCallback()
        callback.cache(10)
        mock_cache_images.assert_called_once_with(10)

    @patch('torchvision.utils.make_grid')
    def test_make_grid(self, mock_grid):
        callback = imaging.ImagingCallback()
        callback.on_batch = lambda _: 10
        callback.make_grid(1, 2, True, 4, True, 6)
        callback.on_batch({})
        mock_grid.assert_called_once_with(10, nrow=1, padding=2, normalize=True, range=4, scale_each=True, pad_value=6)

    def test_process(self):
        callback = imaging.ImagingCallback()
        callback.on_batch = MagicMock()
        handler = MagicMock()
        callback = callback.with_handler(handler)
        callback.transform = MagicMock()
        state = 'state'
        callback.process(state)
        self.assertTrue(callback.transform.call_count == 1)
        handler.assert_called_once_with(callback.transform()[None], 0, 'state')
        self.assertTrue(callback.transform().dim.call_count == 1)

    @patch('torchbearer.callbacks.imaging.imaging._to_visdom')
    @patch('torchbearer.callbacks.imaging.imaging._to_tensorboard')
    @patch('torchbearer.callbacks.imaging.imaging._to_pyplot')
    @patch('torchbearer.callbacks.imaging.imaging._to_file')
    def test_simple_methods(self, mock_to_file, mock_to_pyplot, mock_to_tensorboard, mock_to_visdom):
        callback = imaging.ImagingCallback()
        self.assertRaises(NotImplementedError, lambda: callback.on_batch('test'))

        callback = imaging.ImagingCallback().to_file('test')
        self.assertTrue(mock_to_file.call_count == 1)

        callback = imaging.ImagingCallback().to_pyplot()
        self.assertTrue(mock_to_pyplot.call_count == 1)

        callback = imaging.ImagingCallback().to_state('test')
        state = {}
        callback._handlers[0][0]('image', 0, state)
        self.assertTrue('test' in state)
        self.assertTrue(state['test'] is 'image')

        callback = imaging.ImagingCallback().to_state(0)
        state = {}
        callback._handlers[0][0]('image', 0, state)
        self.assertTrue(0 in state)
        self.assertTrue(state[0] is 'image')

        callback = imaging.ImagingCallback().to_tensorboard()
        self.assertTrue(mock_to_tensorboard.call_count == 1)

        callback = imaging.ImagingCallback().to_visdom()
        self.assertTrue(mock_to_visdom.call_count == 1)

    @patch('torchbearer.callbacks.imaging.imaging._to_file')
    def test_index(self, mock_to_file):
        callback = imaging.ImagingCallback().to_file('test', index=10)
        image = torch.zeros(11, 1, 1, 1)
        callback.on_batch = lambda _: image
        callback.process('state')
        mock_to_file().assert_called_once_with(ANY, 10, 'state')
        self.assertTrue((mock_to_file().call_args_list[0][0][0] == image[10]).all())
        mock_to_file().reset_mock()

        callback = imaging.ImagingCallback().to_file('test', index=[2, 5, 10])
        image = torch.rand(11, 1, 1, 1)
        callback.on_batch = lambda _: image
        callback.process('state')
        self.assertTrue(mock_to_file().call_count == 3)
        self.assertTrue((mock_to_file().call_args_list[0][0][0] == image[2]).all())
        self.assertTrue((mock_to_file().call_args_list[1][0][0] == image[5]).all())
        self.assertTrue((mock_to_file().call_args_list[2][0][0] == image[10]).all())

    @patch('torchbearer.callbacks.imaging.imaging._to_file')
    def test_dims(self, mock_to_file):
        callback = imaging.ImagingCallback().to_file('test')
        image = torch.rand(1, 1, 1)
        callback.on_batch = lambda _: image
        callback.process('state')
        mock_to_file().assert_called_once_with(ANY, 0, 'state')
        self.assertTrue((mock_to_file().call_args_list[0][0][0] == image).all())
        self.assertTrue(mock_to_file().call_args_list[0][0][0].dim() == 3)

    def test_on_train(self):
        callback = imaging.ImagingCallback()
        mock = MagicMock()
        callback.on_step_training = mock
        callback.process = MagicMock()
        callback = callback.on_train()
        callback.on_step_training('state')
        mock.assert_called_once_with('state')
        callback.process.assert_called_once_with('state')

    def test_on_val(self):
        callback = imaging.ImagingCallback()
        mock = MagicMock()
        callback.on_step_validation = mock
        callback.process = MagicMock()
        callback = callback.on_val()
        state = {torchbearer.DATA: torchbearer.TEST_DATA}
        callback.on_step_validation(state)
        mock.assert_called_once_with(state)
        self.assertTrue(callback.process.call_count == 0)

        mock.reset_mock()
        callback.process.reset_mock()

        state = {torchbearer.DATA: torchbearer.VALIDATION_DATA}
        callback.on_step_validation(state)
        mock.assert_called_once_with(state)
        callback.process.assert_called_once_with(state)

    def test_on_test(self):
        callback = imaging.ImagingCallback()
        mock = MagicMock()
        callback.on_step_validation = mock
        callback.process = MagicMock()
        callback = callback.on_test()
        state = {torchbearer.DATA: torchbearer.VALIDATION_DATA}
        callback.on_step_validation(state)
        mock.assert_called_once_with(state)
        self.assertTrue(callback.process.call_count == 0)

        mock.reset_mock()
        callback.process.reset_mock()

        state = {torchbearer.DATA: torchbearer.TEST_DATA}
        callback.on_step_validation(state)
        mock.assert_called_once_with(state)
        callback.process.assert_called_once_with(state)


class TestCachingImagingCallback(TestCase):
    def test_main(self):
        callback = imaging.CachingImagingCallback(key='my_key', num_images=5)
        self.assertRaises(NotImplementedError, lambda: callback.on_cache('', {}))

        callback.on_cache = MagicMock()
        callback.on_cache.return_value = 'image'

        state = {'my_key': torch.ones(10, 3, 2, 2), torchbearer.BATCH: 0}
        callback.on_batch(state)
        self.assertTrue(callback.on_cache.call_args[0][0].sum() == 60)
        self.assertTrue(callback.on_cache.call_args[0][1] == state)

    def test_multi_batch(self):
        callback = imaging.CachingImagingCallback(key='my_key', num_images=25)
        self.assertRaises(NotImplementedError, lambda: callback.on_cache('', {}))

        callback.on_cache = MagicMock()
        callback.on_cache.return_value = 'image'

        state = {'my_key': torch.ones(10, 3, 2, 2), torchbearer.BATCH: 0}
        callback.on_batch(state)
        state[torchbearer.BATCH] = 1
        callback.on_batch(state)
        state[torchbearer.BATCH] = 2
        callback.on_batch(state)
        self.assertTrue(callback.on_cache.call_args[0][0].sum() == 300)
        self.assertTrue(callback.on_cache.call_args[0][1] == state)

    def test_multi_epoch(self):
        callback = imaging.CachingImagingCallback(key='my_key', num_images=5)
        self.assertRaises(NotImplementedError, lambda: callback.on_cache('', {}))

        callback.on_cache = MagicMock()
        callback.on_cache.return_value = 'image'

        state = {'my_key': torch.ones(10, 3, 2, 2), torchbearer.BATCH: 0}
        callback.on_batch(state)
        self.assertTrue(callback.on_cache.call_args[0][0].sum() == 60)
        self.assertTrue(callback.on_cache.call_args[0][1] == state)
        callback.on_cache.reset_mock()
        callback.on_end_epoch({})
        callback.on_batch(state)
        self.assertTrue(callback.on_cache.call_args[0][0].sum() == 60)
        self.assertTrue(callback.on_cache.call_args[0][1] == state)


class TestMakeGrid(TestCase):
    @patch('torchvision.utils.make_grid')
    def test_main(self, mock_grid):
        mock_grid.return_value = 10

        callback = imaging.MakeGrid(key='x', num_images=18, nrow=9, padding=3, normalize=True, norm_range='tmp',
                                    scale_each=True, pad_value=1)

        res = callback.on_cache('test', {})
        mock_grid.assert_called_once_with('test', nrow=9, padding=3, normalize=True, range='tmp', scale_each=True,
                                          pad_value=1)
        self.assertTrue(res == 10)


class TestFromState(TestCase):
    def test_main(self):
        callback = imaging.FromState('test')

        self.assertTrue(callback.on_batch({torchbearer.EPOCH: 0, 'test': 1}) == 1)
        self.assertTrue(callback.on_batch({torchbearer.EPOCH: 1, 'testing': 1}) is None)



from torchbearer import Callback
from .callbacks import *
from .unpack_state import *
from .cutout import Cutout, RandomErase, CutMix
from .lsuv import LSUV
from .checkpointers import *
from .csv_logger import *
from .early_stopping import *
from .gradient_clipping import *
from .printer import ConsolePrinter, Tqdm
from .tensor_board import TensorBoard, TensorBoardImages, TensorBoardProjector, TensorBoardText
from .terminate_on_nan import *
from .torch_scheduler import *
from .weight_decay import *
from .aggregate_predictions import *
from .decorators import *
from .live_loss_plot import LiveLossPlot
from . import init
from . import imaging
from .pycm import PyCM
from .mixup import Mixup, MixupAcc
from .sample_pairing import SamplePairing
from .label_smoothing import LabelSmoothingRegularisation
from .between_class import BCPlus



import torchbearer
from torchbearer import Callback

import torch


def _to_file(filename):
    from PIL import Image

    def handler(image, index, _):
        ndarr = image.mul(255).clamp(0, 255).byte().permute(1, 2, 0).cpu().numpy()
        im = Image.fromarray(ndarr)
        im.save(filename.format(index=str(index)))

    return handler


def _to_pyplot(title=None, show=True):
    import matplotlib.pyplot as plt

    def handler(image, index, _):
        ndarr = image.mul(255).clamp(0, 255).byte().permute(1, 2, 0).cpu().numpy()
        plt.imshow(ndarr)
        if title is not None:
            plt.title(title.format(index=str(index)))
        plt.axis('off')

        if show:
            plt.show()

    return handler


def _to_tensorboard(name='Image', log_dir='./logs', comment='torchbearer'):
    import torchbearer.callbacks.tensor_board as tb
    import os
    log_dir = os.path.join(log_dir, comment)

    def handler(image, index, state):
        writer = tb.get_writer(log_dir, _to_tensorboard)
        writer.add_image(name.format(index=str(index)), image.clamp(0, 1), state[torchbearer.EPOCH])
        tb.close_writer(log_dir, _to_tensorboard)

    return handler


def _to_visdom(name='Image', log_dir='./logs', comment='torchbearer', visdom_params=None):
    import torchbearer.callbacks.tensor_board as tb
    import os
    log_dir = os.path.join(log_dir, comment)

    def handler(image, index, state):
        writer = tb.get_writer(log_dir, _to_visdom, visdom=True, visdom_params=visdom_params)
        writer.add_image(name.format(index=str(index)) + '_' + str(state[torchbearer.EPOCH]), image.clamp(0, 1), state[torchbearer.EPOCH])
        tb.close_writer(log_dir, _to_visdom)

    return handler


def _cache_images(num_images):
    cache = {'images': None, 'done': False}

    def decorator(fun):
        def step(state):
            if state[torchbearer.BATCH] == 0:
                cache['done'] = False

            if not cache['done']:
                data = fun(state)

                if cache['images'] is None:
                    remaining = num_images if num_images < data.size(0) else data.size(0)

                    cache['images'] = data[:remaining]
                else:
                    remaining = num_images - cache['images'].size(0)

                    if remaining > data.size(0):
                        remaining = data.size(0)

                    cache['images'] = torch.cat((cache['images'], data[:remaining]), dim=0)

                if cache['images'].size(0) >= num_images:
                    res = cache['images']
                    cache['done'] = True
                    cache['images'] = None
                    return res
        return step
    return decorator


class ImagingCallback(Callback):
    """The :class:`ImagingCallback` provides a generic interface for callbacks which yield images that should be sent to
    a file, tensorboard, visdom etc. without needing bespoke code. This allows the user to easily define custom
    visualisations by only writing the code to produce the image.

    Args:
        transform (callable, optional): A function/transform that  takes in a Tensor and returns a transformed version.
            This will be applied to the image before it is sent to output.
    """
    def __init__(self, transform=None):
        self._handlers = []
        self.transform = (lambda img: img) if transform is None else transform

    def on_batch(self, state):
        raise NotImplementedError

    def process(self, state):
        img = self.on_batch(state)
        if img is not None:
            img = self.transform(img)
            for handler, index in self._handlers:
                if img.dim() == 3:
                    img = img.unsqueeze(0)
                rng = range(img.size(0)) if index is None else index
                try:
                    for i in rng:
                        handler(img[i], i, state)
                except TypeError:
                    handler(img[rng], rng, state)

    def on_train(self):
        """Process this callback for training batches

        Returns:
            ImagingCallback: self
        """
        _old_step_training = self.on_step_training

        def wrapper(state):
            _old_step_training(state)
            self.process(state)

        self.on_step_training = wrapper
        return self

    def on_val(self):
        """Process this callback for validation batches

        Returns:
            ImagingCallback: self
        """
        _old_step_validation = self.on_step_validation

        def wrapper(state):
            _old_step_validation(state)
            if state[torchbearer.DATA] is torchbearer.VALIDATION_DATA:
                self.process(state)

        self.on_step_validation = wrapper
        return self

    def on_test(self):
        """Process this callback for test batches

        Returns:
            ImagingCallback: self
        """
        _old_step_validation = self.on_step_validation

        def wrapper(state):
            _old_step_validation(state)
            if state[torchbearer.DATA] is torchbearer.TEST_DATA:
                self.process(state)

        self.on_step_validation = wrapper
        return self

    def with_handler(self, handler, index=None):
        """Append the given output handler to the list of handlers

        Args:
            handler: A function of image and state which stores the given image in some way
            index (int or list or None): If not None, only apply the handler on this index / list of indices

        Returns:
            ImagingCallback: self
        """
        self._handlers.append((handler, index))
        return self

    def to_file(self, filename, index=None):
        """Send images from this callback to the given file

        Args:
            filename (str): The filename to store the image to
            index (int or list or None): If not None, only apply the handler on this index / list of indices

        Returns:
            ImagingCallback: self
        """
        return self.with_handler(_to_file(filename), index=index)

    def to_pyplot(self, title=None, show=True, index=None):
        """Show images from this callback with pyplot

        Args:
            title (str or None): If not None, plt.title will be called with the given string
            show (bool): If True (default), show will be called after each image is plotted
            index (int or list or None): If not None, only apply the handler on this index / list of indices

        Returns:
            ImagingCallback: self
        """
        return self.with_handler(_to_pyplot(title=title, show=show), index=index)

    def to_state(self, keys, index=None):
        """Put images from this callback in state with the given key

        Args:
            keys (StateKey or list[StateKey]): The state key or keys to use for the images
            index (int or list or None): If not None, only apply the handler on this index / list of indices

        Returns:
            ImagingCallback: self
        """
        if str(keys) == keys:
            keys = [keys]

        try:
            _ = (key for key in keys)
        except TypeError:
            keys = [keys]

        def handler(img, i, state):
            state[keys[i]] = img
        return self.with_handler(handler, index=index)

    def to_tensorboard(self, name='Image', log_dir='./logs', comment='torchbearer', index=None):
        """Direct images from this callback to tensorboard with the given parameters

        Args:
            name (str): The name of the image
            log_dir (str): The tensorboard log path for output
            comment (str): Descriptive comment to append to path
            index (int or list or None): if not None, only apply the handler on this index / list of indices

        Returns:
            ImagingCallback: self
        """
        return self.with_handler(_to_tensorboard(name=name, log_dir=log_dir, comment=comment), index=index)

    def to_visdom(self, name='Image', log_dir='./logs', comment='torchbearer', visdom_params=None, index=None):
        """Direct images from this callback to visdom with the given parameters

        Args:
            name (str): The name of the image
            log_dir (str): The visdom log path for output
            comment (str): Descriptive comment to append to path
            visdom_params (:class:`.VisdomParams`): Visdom parameter settings object, uses default if None
            index (int or list or None): if not None, only apply the handler on this index / list of indices

        Returns:
            ImagingCallback: self
        """
        return self.with_handler(_to_visdom(name=name, log_dir=log_dir, comment=comment, visdom_params=visdom_params), index=index)

    def cache(self, num_images):
        """Cache images **before** they are passed to handlers. Once per epoch, a single cache will be returned,
        containing the first `num_images` to be returned.

        Args:
            num_images (int): The number of images to cache

        Returns:
            ImagingCallback: self
        """
        self.on_batch = _cache_images(num_images)(self.on_batch)
        return self

    def make_grid(self, nrow=8, padding=2, normalize=False, norm_range=None, scale_each=False, pad_value=0):
        """Use `torchvision.utils.make_grid` to make a grid of the images being returned by this callback. Recommended
        for use alongside `cache`.

        Args:
            nrow: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
            padding: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
            normalize: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
            norm_range: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
            scale_each: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
            pad_value: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_

        Returns:
            ImagingCallback: self
        """
        import torchvision.utils as utils

        def decorator(func):
            def wrapper(state):
                cache = func(state)
                if cache is not None:
                    return utils.make_grid(cache, nrow=nrow, padding=padding, normalize=normalize, range=norm_range,
                                           scale_each=scale_each, pad_value=pad_value)
            return wrapper

        self.on_batch = decorator(self.on_batch)
        return self


class FromState(ImagingCallback):
    """The :class:`FromState` callback is an :class:`ImagingCallback` which retrieves and image from state when called.
    The number of times the function is called can be controlled with a provided decorator (once_per_epoch, only_if
    etc.)

    Args:
        key (StateKey): The :class:`.StateKey` containing the image (tensor of size [c, w, h])
        transform (callable, optional): A function/transform that  takes in a Tensor and returns a transformed version.
            This will be applied to the image before it is sent to output.
        decorator: A function which will be used to wrap the callback function. once_per_epoch by default
    """
    def __init__(self, key, transform=None, decorator=None):
        super(FromState, self).__init__(transform=transform)
        self.key = key

        if decorator is not None:
            self.on_batch = decorator(self.on_batch)

    def on_batch(self, state):
        try:
            return state[self.key]
        except KeyError:
            return None


class CachingImagingCallback(FromState):
    """The :class:`CachingImagingCallback` is an :class:`ImagingCallback` which caches batches of images from the given
    state key up to the required amount before passing this along with state to the implementing class, once per epoch.

    Args:
        key (StateKey): The :class:`.StateKey` containing image data (tensor of size [b, c, w, h])
        transform (callable, optional): A function/transform that  takes in a Tensor and returns a transformed version.
            This will be applied to the image before it is sent to output.
        num_images: The number of images to cache
    """
    def __init__(self,
                 key=torchbearer.INPUT,
                 transform=None,
                 num_images=16):
        super(CachingImagingCallback, self).__init__(key=key, transform=transform, decorator=_cache_images(num_images))

        def decorator(func):
            def wrapper(state):
                res = func(state)
                if res is not None:
                    return self.on_cache(res, state)
            return wrapper

        self.on_batch = decorator(self.on_batch)

    def on_cache(self, cache, state):
        """This method should be implemented by the overriding class to return an image from the cache.

        Args:
            cache (tensor): The collected cache of size (num_images, C, W, H)
            state (dict): The trial state dict

        Returns:
            The processed image
        """
        raise NotImplementedError


class MakeGrid(CachingImagingCallback):
    """The :class:`MakeGrid` callback is a :class:`CachingImagingCallback` which calls make grid on the cache with the
    provided parameters.

    Args:
        key (StateKey): The :class:`.StateKey` containing image data (tensor of size [b, c, w, h])
        transform (callable, optional): A function/transform that  takes in a Tensor and returns a transformed version.
            This will be applied to the image before it is sent to output.
        num_images: The number of images to cache
        nrow: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
        padding: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
        normalize: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
        norm_range: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
        scale_each: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
        pad_value: See `torchvision.utils.make_grid <https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid>`_
    """
    def __init__(self,
                 key=torchbearer.INPUT,
                 transform=None,
                 num_images=16,
                 nrow=8,
                 padding=2,
                 normalize=False,
                 norm_range=None,
                 scale_each=False,
                 pad_value=0):
        super(MakeGrid, self).__init__(transform=transform, num_images=num_images, key=key)
        self.key = key
        self.num_images = num_images
        self.nrow = nrow
        self.padding = padding
        self.normalize = normalize
        self.norm_range = norm_range
        self.scale_each = scale_each
        self.pad_value = pad_value

    def on_cache(self, cache, state):
        import torchvision.utils as utils
        return utils.make_grid(
            cache,
            nrow=self.nrow,
            padding=self.padding,
            normalize=self.normalize,
            range=self.norm_range,
            scale_each=self.scale_each,
            pad_value=self.pad_value
        )
